{***********************************************************

Project:    C-evo External Map Generator
Copyright:  1999-2023 P Blackman
License:    GPLv3+

City Start Positions & Deadlands
Deadlands should be within a viable city site

***********************************************************}


function tMap.citysquares : Integer;
begin
    result := Citylist.Count;
end;


procedure tMap.GetCityNeighbour (W, H, N : Integer; out Wt, Ht : Integer);
var O : Integer;
begin
    Wt := W;
    Ht := H;
    O := Integer (not odd (H)); { Even rows need an offset }

    { Get Co-Ordinates for this Neighbour }
    case N of
    0: ;   { my city square }
    1,2:
        begin
            Ht := H - 3;
            Wt := W - 2 + N + O;
        end;
    3,4,5:
        begin
            Ht := H - 2;
            Wt := W - 4 + N;
        end;
    6,7,8,9:
        begin
            Ht := H - 1;
            Wt := W - 7 + N + O;
        end;
    10,11:
        begin
            Ht := H;
            Wt := W + 2*(11-N) -1;  {-1 or +1}
        end;
    12,13,14,15:
        begin
            Ht := H + 1;
            Wt := W - 14 + N + O;
        end;
    16,17,18:
        begin
            Ht := H + 2;
            Wt := W - 17 + N;
        end;
    19,20:  begin
            Ht := H + 3;
            Wt := W - 20 + N + O;
        end;
    end;

    Wt := WrapCheck (Wt);
end;


procedure tMap.DumpCity (W, H: Integer);
var N, Wt, Ht : Integer;
begin
    for N := 0 to 20 do
    begin
        GetCityNeighbour (W, H, N, Wt, Ht);
        Assert ((Wt > 0) and (Wt <= Width),  'Invalid Width ' +IntToStr(Wt));
        Assert ((Ht > 0) and (Ht <= Height), 'Invalid Height '+IntToStr(Ht));

        Write (TerrainChar(Getterrain(Wt,Ht)));
    end;
end;


function tMap.CoastalCity (const W,H : Integer) : Boolean;
var T,Wn,Hn : Integer;
    Landlocked : Boolean;
begin
    LandLocked := True;
    T := 0;
    while landlocked and (T < 8) do
    begin
        inc (T);

        GetNeighbour (W,H,T,Wn,Hn);

        If (Hn > 0) and (Hn <= Height) then
            landlocked := not (GetTerrain (Wn, Hn) = Coast);
    end;
    result := not LandLocked;
end;


procedure tMap.TileResource (const W,H : Integer;
                                Coastal : Boolean;
                                out F,M,Tr : Integer;
                                out PosDeadLands : Boolean);
var T : tTerrain;
begin
    F := 0;
    M := 0;
    T := GetTerrain (W,H);
    Tr := 0; {Trade}

    { Food where two or more per tile with irregation }
    If (T in [Grass, Prairie, Tundra, Hills]) then
    case T of
        Grass:  F := 4;
        Prairie:
            begin
                F := 2;
                M := 1;
            end;
        Tundra,
        Hills:  F := 2;
    end;

    { Trade }
    If (T in [Grass, Prairie, Tundra]) then
        Tr := 2 + Integer(Tiles[W,H].River)
    else
    if Coastal and (T = Coast) then
        Tr := 3;

    { Bonus resources. Only first type usable in early game }
    If (GetBonus (W,H) = EarlyBonus) and (T in [Grass, Coast, Desert, Prairie, Forest, Jungle, Arctic, Tundra, Hills]) then
    case T of
        Grass:
            begin
                { plains }
                F := 3;
                M := 1;
            end;
        Coast:
            if Coastal then
                F := 6; { Add one more for harbour }
        Desert:
            begin
                F := 3;
                M := 1;
            end;
        Prairie:
            begin
                F := 4;
                M := 1;
            end;
        Forest,
        Jungle: F := 3;
        Tundra: Tr:= Tr+5;
        Arctic:
            begin
                F  := 3;
                M  := 3;
                Tr := 4;
            end;
        Hills : Tr := Tr+4;
    end;

    PosDeadLands := (getBonus (W,H) = Nothing) and (T in [Desert, Arctic, Mountain]);
end;


function tMap.StartScore (W,H : Integer; out DeadLand : Boolean) : Integer;
var Wt,Ht,T,F,E,M,Ft,Mt,Tr,Trt : Integer;
    Coastal,
    DL : Boolean;
begin
    F := 0;
    E := 0;
    M := 0;
    Tr:= 0;
    DeadLand := False;
    Coastal := CoastalCity (W,H);

    for T := 0 to 20 do
    begin
        GetCityNeighbour (W, H, T, Wt, Ht);

        If (Ht > 0) and (Ht <= Height) then
        begin
            TileResource (Wt, Ht, Coastal, Ft, Mt, Trt, DL);
            F  := F + Ft;
            M  := M + Mt;
            Tr := Tr+ Trt;
            If Ft > 2 then
                E := E + Ft-2;  { Extra Food above two per tile}

            DeadLand := DeadLand or DL;
        end;
     end;

     { Bonus for coastal city }
     result := 2*Integer (Coastal) + trunc (sqrt(sqrt(E*F*M*Tr))); {Reduce score to convenient number}
end;

Procedure tMap.AddCity (W,H,S : Integer; T : tTerrain; DL : Boolean);
var ptr : tTileptr;
begin
    New (ptr);
    with ptr^ do
    begin
        Wid   := W;
        Hgt   := H;
        Score := S;
        Tile  := T;
        Cull  := False;
        DeadLand := DL;
    end;
    Citylist.add (ptr);
end;

function tMap.Dist (W1,H1,W2,H2 : Integer) : Integer;
var Wd, Hd : Integer;
begin
    Hd := Abs (H1-H2) div 2;
    Wd := Abs (W1-W2);

    { Round World, Horizontal distance cannot be more than half the width }
    if Wd > Width Div 2 then
        Wd := Width - Wd;

    result := Hd+Wd;
end;

procedure tMap.CullCitySites;
var C1, C2 : Longint;
    D : Integer;
    Cp1, Cp2 : tTileptr;
    Proximity : Boolean;
begin
    For C1 := 0 to Citylist.Count-1 do
    begin
        Proximity := False;
        C2 := C1;
        Cp1 := Citylist.Items[C1];
        While not Proximity and (C2 > 0) do
        begin
            Dec (C2);
            Cp2 := Citylist.Items[C2];
            D  := Dist (Cp1^.Wid, Cp1^.Hgt, Cp2^.Wid, Cp2^.Hgt);

            Proximity := not Cp2^.cull and (D < MinCityDist);
            If Proximity then
                 Cp1^.cull := True;
         end;
    end;

    For C1 := 0 to Citylist.Count-1 do
    begin
        Cp1 := Citylist.Items[C1];
        If Cp1^.cull then
        begin
            Dispose (Cp1);
            Citylist.Items[C1] := nil;
        end
    end;
    CityList.Pack;
end;

function tMap.StartPosition (W, H : Integer) : Boolean;
begin
    result := Tiles [W,H].StartPos;
end;

procedure tMap.SetStartPosition (W, H : Integer);
begin
    Tiles [W,H].StartPos := True;
end;


function tMap.CheckDeadLand (const C : tTileptr; T : tTerrain; out DLW, DLH : Integer) : Boolean;
var N, W, H : Integer;
begin
    DLW := 0;
    DLH := 0;

    for N := 1 to 20 do
    begin
        GetCityNeighbour (C^.Wid, C^.Hgt, N, W, H);
        If T = GetTerrain(W,H) then
        begin
            DLW := W;
            DLH := H;
        end;
    end;

    result := DLW <> 0;
end;

procedure tMap.SetDeadTile(D,W,H : Integer);
Var DT : tSpecial;
begin
    case D of
        { can build a space ship with first three only }
        1,4,7:    DT := Cobalt;
        2,5,8:    DT := Uranium;
        3,6,9:    DT := Mercury;
      otherwise DT := DeadLands;
    end;

    SetTerrain (W, H, Desert, False);
    Tiles [W, H].Special := DT;
end;

procedure tMap.SetDeadLand (const C : tTileptr; Cn, D : Integer);
var W,H : Integer;
begin
    If CheckDeadLand (C, Desert, W, H) then
        SetDeadTile (D,W,H)
    else
    If CheckDeadLand (C, Arctic, W, H) then
        SetDeadTile (D,W,H)
    else
    If CheckDeadLand (C, Mountain, W, H) then
        SetDeadTile (D,W,H)
    else
    begin
        Writeln ('Logic error in DeadLands ',Cn, D:3);
        Halt (1);
    end;
end;

procedure tMap.AllocateDeadLands;
var C,D : Integer;
    CitySite : tTileptr;
begin
    C := 0;
    D := 0;

    While (C < Citylist.Count) and (D < NumDeadlands)do
    begin
        CitySite := Citylist.Items[C];

        If CitySite^.DeadLand then
        begin
            inc (D);
            SetDeadLand (CitySite, C, D);

            if Citylist.Count > nPl then
            begin
                { Ensure site not used as a start position }
                Dispose (CitySite);
                Citylist.Items[C] := nil;
            end;
        end;

        inc (C);
    end;

    CityList.Pack;
end;


procedure tMap.PlaceStartPositions;
var C, W, H, S : Integer;
    DL : Boolean;
    T : tTerrain;
    Cs : tTileptr;
begin
    for W := 1 to Width do
        for H := 4 to Height -3 do
        begin
            {Check each Tile for possible city site}
            T := GetTerrain (W,H);
            if T in [Grass, Prairie, Hills, Tundra] then
            begin
                S := StartScore (W,H,DL);
                If S >= MinCityScore then
                    AddCity (W,H,S,T,DL);
            end;
        end;

    { Sort City Sites by food value }
    Citylist.sort (@CitySort);

    { Cull overlapping entries }
    CullCitySites;

    AllocateDeadLands;

    { Allocate Starting Positions }
    for C := 1 to min (NumStartPos, Citylist.Count) do
    begin
        Cs := Citylist.Items[C-1];
        with Cs^ do
            Tiles [Wid,Hgt].StartPos := True;
    end;
end;


procedure tMap.CountCitySites;
var W, H, S : Integer;
    DL : Boolean;
    T : tTerrain;

begin
    for W := 1 to Width do
        for H := 4 to Height -3 do
        begin
            {Check each Tile for possible city site}
            T := GetTerrain (W,H);
            if T in [Grass, Prairie, Hills, Tundra] then
            begin
                S := StartScore (W,H,DL);
                If S >= MinCityScore then
                    AddCity (W,H,S,T,DL);
            end;
        end;

    { Sort City Sites by food value }
    Citylist.sort (@CitySort);

    { Cull overlapping entries }
    CullCitySites;
end;
